# TODO: This file needs to allow explicitly the following:
# Run a scan on a particular path with any file selection criteria
# Load results of any previous scan
# Automatically find canidate results to load.
